# generic source class:
from .sph_source import SPHSource

# toolkit-specific source classes:
from .so_source import SOSource
from .swiftgalaxy_source import SWIFTGalaxySource

# simulation-specific source classes:
from .colibre_source import ColibreSource
from .eagle_source import EAGLESource
from .magneticum_source import MagneticumSource
from .simba_source import SimbaSource
from .tng_source import TNGSource



import numpy as np
import astropy.units as U

# copied from github.com/kyleaoman/kyleaoman_utilities/kyleaoman_utilities/
# commit-id 63dc535c5de1a0dc3caf09f98a33c9d09aa0eddc
# Note: No git-based solution (e.g. via submodules) seems practical to include
# selected files from external repositories; a direct copy is included here
# to produce a self-contained package.


def L_align(xyz, vxyz, m, frac=0.3, saverot=None, Laxis="z"):
    transposed = False
    if xyz.ndim != 2:
        raise ValueError(
            "L_align: cannot guess coordinate axis for input with" " ndim != 2."
        )
    elif (xyz.shape[0] == 3) and (xyz.shape[1] == 3):
        raise ValueError(
            "L_align: cannot guess coordinate axis for input with" " shape (3, 3)."
        )
    elif xyz.shape[1] == 3:
        xyz = xyz.T
        vxyz = vxyz.T
        transposed = True
    elif (xyz.shape[0] != 3) and (xyz.shape[1] != 3):
        raise ValueError(
            "L_align: coordinate array shape "
            + str(xyz.shape)
            + " invalid (one dim must be 3)."
        )

    rsort = np.argsort(np.sum(np.power(xyz, 2), axis=0), kind="quicksort")
    p = m[np.newaxis] * vxyz
    L = np.cross(xyz, p, axis=0)
    p = p[:, rsort]
    L = L[:, rsort]
    m = m[rsort]
    mcumul = np.cumsum(m) / np.sum(m)
    Nfrac = np.argmin(np.abs(mcumul - frac))
    Nfrac = np.max([Nfrac, 100])  # use a minimum of 100 particles
    Nfrac = np.min([Nfrac, len(m)])  # unless this exceeds particle count
    p = p[:, :Nfrac]
    L = L[:, :Nfrac]
    Ltot = np.sqrt(np.sum(np.power(np.sum(L, axis=1), 2)))
    Lhat = np.sum(L, axis=1) / Ltot
    zhat = Lhat / np.sqrt(np.sum(np.power(Lhat, 2)))  # normalized
    xaxis = np.array([1.0, 1.0, 1.0]) / np.sqrt(3)  # default unlikely Laxis
    if (zhat == xaxis).all():
        raise RuntimeError(
            "Angular momentum exactly aligned with arbitrarily chosen vector, L_align"
            " failed."
        )
    xhat = xaxis - xaxis.dot(zhat) * zhat
    xhat = xhat / np.sqrt(np.sum(np.power(xhat, 2)))  # normalized
    yhat = np.cross(zhat, xhat)  # guarantees right-handedness

    rotmat = np.vstack((xhat, yhat, zhat)).to(U.dimensionless_unscaled).value
    if Laxis == "z":
        pass
    elif Laxis == "y":
        rotmat = np.roll(rotmat, 2, axis=0)
    elif Laxis == "x":
        rotmat = np.roll(rotmat, 1, axis=0)
    else:
        raise ValueError("L_align: Laxis must be one of 'x', 'y' or 'z'.")

    if transposed:
        rotmat = rotmat.T
    if saverot is not None:
        np.save(saverot, rotmat)

    return rotmat



import pytest
import numpy as np
from astropy import units as U
from martini.beams import GaussianBeam
from martini import DataCube


class TestGaussianBeam:
    def test_Gaussian_beam_size(self):
        """
        Check that we produce a square beam array large enough to contain it.
        """
        px_size = 1 * U.arcsec
        bmaj = 20 * U.arcsec
        bmin = 10 * U.arcsec
        bpa = 0 * U.deg
        truncate = 4
        d = DataCube(
            n_px_x=1,
            n_px_y=1,
            n_channels=1,
            px_size=px_size,
            channel_width=1 * U.km / U.s,
        )
        b = GaussianBeam(bmaj=bmaj, bmin=bmin, bpa=bpa, truncate=truncate)
        b.init_kernel(d)
        halfsize = (
            int(np.ceil((bmaj * truncate / px_size).to_value(U.dimensionless_unscaled)))
            + 1
        )
        expected_size = 2 * halfsize + 1
        assert b.kernel.shape == (expected_size, expected_size)

    @pytest.mark.parametrize("bpa", (0 * U.deg, 90 * U.deg))
    def test_Gaussian_beam_rotation(self, bpa):
        """
        Check that the beam amplitude is equal along the ellipse traced by the beam shape.
        """
        px_size = 1 * U.arcsec
        bmaj = 20 * U.arcsec
        bmin = 10 * U.arcsec
        truncate = 4
        d = DataCube(
            n_px_x=1,
            n_px_y=1,
            n_channels=1,
            px_size=px_size,
            channel_width=1 * U.km / U.s,
        )
        b = GaussianBeam(bmaj=bmaj, bmin=bmin, bpa=bpa, truncate=truncate)
        b.init_kernel(d)
        mid = b.kernel.shape[0] // 2
        if bpa == 0 * U.deg:
            assert U.isclose(
                b.kernel[mid, mid + 5],
                b.kernel[mid + 10, mid],
                rtol=1e-2,
            )
        elif bpa == 90 * U.deg:
            assert U.isclose(
                b.kernel[mid, mid + 10],
                b.kernel[mid + 5, mid],
                rtol=1e-2,
            )

    def test_Gaussian_beam_integral(self):
        """
        Check that the beam integrates to 1.0.
        """
        px_size = 1 * U.arcsec
        bmaj = 10 * U.arcsec
        bmin = 10 * U.arcsec
        truncate = 4
        d = DataCube(
            n_px_x=1,
            n_px_y=1,
            n_channels=1,
            px_size=px_size,
            channel_width=1 * U.km / U.s,
        )
        b = GaussianBeam(bmaj=bmaj, bmin=bmin, bpa=0 * U.deg, truncate=truncate)
        b.init_kernel(d)
        assert U.isclose(np.sum(b.kernel), 1.0 * U.dimensionless_unscaled)



import io
import os
import numpy as np
import astropy.units as U
import astropy.constants as C
from .sph_source import SPHSource
from ..sph_kernels import _CubicSplineKernel, find_fwhm


def api_get(path, params=None, api_key=None):
    import requests

    baseUrl = "http://www.tng-project.org/api/"
    r = requests.get(f"{baseUrl}/{path}", params=params, headers={"api-key": api_key})
    r.raise_for_status()
    if r.headers["content-type"] == "application/json":
        return r.json()
    return r


def cutout_file(simulation, snapNum, haloID):
    return f"martini-cutout-{simulation}-{snapNum}-{haloID}.hdf5"


class TNGSource(SPHSource):
    """
    Class abstracting HI sources for use with IllustrisTNG simulations.

    If used in the IllustrisTNG JupyterLab environment
    (https://www.tng-project.org/data/lab/), files on disk are accessed directly.
    Otherwise, particles for the galaxy of interest are automatically retrieved using
    the TNG web API (https://www.tng-project.org/data/docs/api/).

    Use of the TNG web API requires an API key: login at
    https://www.tng-project.org/users/login/ or register at
    https://www.tng-project.org/users/register/ then obtain your API
    key from https://www.tng-project.org/users/profile/ and pass to TNGSource as the
    api_key parameter. An API key is not required if logged into the TNG JupyterLab.

    Parameters
    ----------
    simulation : str
        Simulation identifier string, for example "TNG100-1", see
        https://www.tng-project.org/data/docs/background/

    snapNum : int
        Snapshot number. In TNG, snapshot 99 is the final output. Note that
        if a 'mini' snapshot (see
        http://www.tng-project.org/data/docs/specifications/#sec1a) is selected then
        some additional approximations are used.

    subID : int
        Subhalo ID of the target object. Note that all particles in the FOF
        group to which the subhalo belongs are used to construct the data cube.
        This avoids strange 'holes' at the locations of other subhaloes in the
        same group, and gives a more realistic treatment of foreground and
        background emission local to the source. An object of interest could be
        found using https://www.tng-project.org/data/search/, for instance. The
        "ID" column in the search results on that page is the subID.

    api_key: str, optional
        Use of the TNG web API requires an API key: login at
        https://www.tng-project.org/users/login/ or register at
        https://www.tng-project.org/users/register/ then obtain your API
        key from https://www.tng-project.org/users/profile/ and provide as a string. An
        API key is not required if logged into the TNG JupyterLab. (Default: None.)

    cutout_dir: str, optional
        Ignored if running on the TNG JupyterLab. Directory in which to search for and
        save cutout files (see documentation at
        https://www.tng-project.org/data/docs/api/ for a description of cutouts). Cutout
        filenames are enforced by MARTINI. If `cutout_dir==None` (the default), then the
        data will always be downloaded. If a `cutout_dir` is provided, it will first be
        searched for the required data. If the data are found, the local copy is used,
        otherwise the data are downloaded and a local copy is saved in `cutout_dir` for
        future use. (Default: None.)

    distance : Quantity, with dimensions of length, optional
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: 3 Mpc.)

    vpeculiar : Quantity, with dimensions of velocity, optional
        Source peculiar velocity, added to the velocity from Hubble's law.
        (Default: 0 km/s.)

    rotation : dict, optional
        Must have a single key, which must be one of `axis_angle`, `rotmat` or
        `L_coords`. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - `axis_angle` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - `rotmat` : A (3, 3) numpy.array specifying a rotation.
        - `L_coords` : A 2-tuple containing an inclination and an azimuthal \
        angle (both Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: identity rotation matrix.)

    ra : Quantity, with dimensions of angle, optional
        Right ascension for the source centroid. (Default: 0 deg.)

    dec : Quantity, with dimensions of angle, optional
        Declination for the source centroid. (Default: 0 deg.)
    """

    def __init__(
        self,
        simulation,
        snapNum,
        subID,
        api_key=None,
        cutout_dir=None,
        distance=3.0 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
    ):
        # optional dependencies for this source class
        import h5py
        from Hdecompose.atomic_frac import atomic_frac

        X_H = 0.76

        full_fields_g = (
            "Masses",
            "Velocities",
            "InternalEnergy",
            "ElectronAbundance",
            "Density",
            "CenterOfMass",
            "GFM_Metals",
        )
        mdi_full = [None, None, None, None, None, None, 0]
        mini_fields_g = (
            "Masses",
            "Velocities",
            "InternalEnergy",
            "ElectronAbundance",
            "Density",
            "Coordinates",
        )

        # are we running on the TNG jupyterlab?
        jupyterlab = os.path.exists("/home/tnguser/sims.TNG")
        if not jupyterlab:
            data_header = dict()
            data_sub = dict()
            data_g = dict()
            from requests import HTTPError

            if api_key is None:
                raise ValueError(
                    "A TNG API key is required: login at "
                    "https://www.tng-project.org/users/login/ or register at "
                    "https://www.tng-project.org/users/register/ then obtain your API "
                    "key from https://www.tng-project.org/users/profile/"
                )
            if cutout_dir is not None:
                grnr_file = os.path.join(
                    cutout_dir,
                    f"martini-cutout-grnr-{simulation}-{snapNum}-{subID}.npy",
                )
                if os.path.exists(grnr_file):
                    haloID = np.load(grnr_file)
                    have_cutout = True
                else:
                    have_cutout = False
            else:
                print("No cutout_dir provided, cutout will be downloaded.")
                have_cutout = False
            if have_cutout:
                # check for an existing local cutout file
                if not os.path.exists(
                    os.path.join(cutout_dir, cutout_file(simulation, snapNum, haloID))
                ):
                    have_cutout = False
                cfname = os.path.join(
                    cutout_dir, cutout_file(simulation, snapNum, haloID)
                )
                print(f"Using local cutout file {os.path.basename(cfname)}")
            if not have_cutout:  # not else because previous if can modify have_cutout
                print("No local cutout found, cutout will be downloaded.")
                sub_api_path = f"{simulation}/snapshots/{snapNum}/subhalos/{subID}"
                sub = api_get(sub_api_path, api_key=api_key)
                haloID = sub["grnr"]
                np.save(grnr_file, haloID)
                data_sub["SubhaloPos"] = np.array([sub[f"pos_{ax}"] for ax in "xyz"])
                data_sub["SubhaloVel"] = np.array([sub[f"vel_{ax}"] for ax in "xyz"])
                cutout_api_path = (
                    f"{simulation}/snapshots/{snapNum}/halos/{haloID}/cutout.hdf5"
                )
                cutout_request = dict(gas=",".join(full_fields_g))
                try:
                    cutout = api_get(
                        cutout_api_path, params=cutout_request, api_key=api_key
                    )
                except HTTPError:
                    cutout_request = dict(gas=",".join(mini_fields_g))
                    cutout = api_get(
                        cutout_api_path, params=cutout_request, api_key=api_key
                    )
                # hold file in memory
                cfname = io.BytesIO(cutout.content)
                if cutout_dir is not None:
                    # write a copy to disk for later use
                    ofile = os.path.join(
                        cutout_dir, cutout_file(simulation, snapNum, haloID)
                    )
                    print(f"Writing downloaded cutout to {ofile}")
                    with open(ofile, "wb") as of:
                        of.write(cutout.content)
                    with h5py.File(ofile, "r+") as of:
                        of.create_group(f"{subID}")
                        of[f"{subID}"].attrs["pos"] = data_sub["SubhaloPos"]
                        of[f"{subID}"].attrs["vel"] = data_sub["SubhaloVel"]
            with h5py.File(cfname, "r") as cf:
                minisnap = "CenterOfMass" not in cf["PartType0"].keys()
                fields_g = mini_fields_g if minisnap else full_fields_g
                data_g = {field: cf["PartType0"][field][()] for field in fields_g}
                if len(data_header) == 0:
                    data_header["HubbleParam"] = cf["Header"].attrs["HubbleParam"]
                    data_header["Redshift"] = cf["Header"].attrs["Redshift"]
                    data_header["Time"] = cf["Header"].attrs["Time"]
                if len(data_sub) == 0:
                    data_sub["SubhaloPos"] = cf[f"{subID}"].attrs["pos"]
                    data_sub["SubhaloVel"] = cf[f"{subID}"].attrs["vel"]
            X_H_g = X_H if minisnap else data_g["GFM_Metals"][:, 0]

        else:
            from ._illustris_tools import (
                loadHeader,
                loadSingle,
                loadSubset,
                getSnapOffsets,
            )

            if cutout_dir is not None:
                print("Running on TNG JupyterLab, cutout_dir ignored.")
            basePath = f"/home/tnguser/sims.TNG/{simulation}/output/"
            data_header = loadHeader(basePath, snapNum)
            data_sub = loadSingle(basePath, snapNum, subhaloID=subID)
            haloID = data_sub["SubhaloGrNr"]
            subset_g = getSnapOffsets(basePath, snapNum, haloID, "Group")
            try:
                data_g = loadSubset(
                    basePath,
                    snapNum,
                    "gas",
                    fields=full_fields_g,
                    subset=subset_g,
                    mdi=mdi_full,
                )
                minisnap = False
            except Exception as exc:
                if ("Particle type" in exc.args[0]) and (
                    "does not have field" in exc.args[0]
                ):
                    data_g.update(
                        loadSubset(
                            basePath,
                            snapNum,
                            "gas",
                            fields=("CenterOfMass",),
                            subset=subset_g,
                            sq=False,
                        )
                    )
                    minisnap = True
                    X_H_g = X_H
                else:
                    raise
            X_H_g = (
                X_H if minisnap else data_g["GFM_Metals"]
            )  # only loaded column 0: Hydrogen

        a = data_header["Time"]
        z = data_header["Redshift"]
        h = data_header["HubbleParam"]
        xe_g = data_g["ElectronAbundance"]
        rho_g = data_g["Density"] * 1e10 / h * U.Msun * np.power(a / h * U.kpc, -3)
        u_g = data_g["InternalEnergy"]  # unit conversion handled in T_g
        mu_g = 4 * C.m_p.to(U.g).value / (1 + 3 * X_H_g + 4 * X_H_g * xe_g)
        gamma = 5.0 / 3.0  # see http://www.tng-project.org/data/docs/faq/#gen4
        T_g = (gamma - 1) * u_g / C.k_B.to(U.erg / U.K).value * 1e10 * mu_g * U.K
        m_g = data_g["Masses"] * 1e10 / h * U.Msun
        # cast to float64 to avoid underflow error
        nH_g = U.Quantity(rho_g * X_H_g / mu_g, dtype=np.float64) / C.m_p
        # In TNG_corrections I set f_neutral = 1 for particles with density
        # > .1cm^-3. Might be possible to do a bit better here, but HI & H2
        # tables for TNG will be available soon anyway.
        fatomic_g = atomic_frac(
            z, nH_g, T_g, rho_g, X_H_g, onlyA1=True, TNG_corrections=True
        )
        mHI_g = m_g * X_H_g * fatomic_g
        try:
            xyz_g = data_g["CenterOfMass"] * a / h * U.kpc
        except KeyError:
            xyz_g = data_g["Coordinates"] * a / h * U.kpc
        vxyz_g = data_g["Velocities"] * np.sqrt(a) * U.km / U.s
        V_cell = (
            data_g["Masses"] / data_g["Density"] * np.power(a / h * U.kpc, 3)
        )  # Voronoi cell volume
        r_cell = np.power(3.0 * V_cell / 4.0 / np.pi, 1.0 / 3.0).to(U.kpc)
        # hsm_g has in mind a cubic spline that =0 at r=h, I think
        hsm_g = 2.5 * r_cell * find_fwhm(_CubicSplineKernel().kernel)
        xyz_centre = data_sub["SubhaloPos"] * a / h * U.kpc
        xyz_g -= xyz_centre
        vxyz_centre = data_sub["SubhaloVel"] * np.sqrt(a) * U.km / U.s
        vxyz_g -= vxyz_centre

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            T_g=T_g,
            mHI_g=mHI_g,
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            hsm_g=hsm_g,
        )
        return



import numpy as np
import astropy.units as U
from ..sph_kernels import _WendlandC6Kernel, find_fwhm
from .sph_source import SPHSource


class MagneticumSource(SPHSource):
    """
    Class abstracting HI sources designed to work with Magneticum snapshot
    + group fies.

    Parameters
    ----------
    snapBase : str
        Path to snapshot file, omitting the portion numbering the snapshot
        pieces, e.g. /path/snap_136.0 becomes /path/snap_136

    haloPosition : array_like, with shape (3, )
        Location of source centre in simulation units. Provide either
        arguments haloPosition, haloVelocity and haloRadius, or arguments
        groupFile, (haloID | subhaloID), not both.

    haloVelocity : array_like, with shape (3, )
        Velocity of halo in the simulation box frame, in simulation units.
        Provide either arguments haloPosition, haloVelocity and
        haloRadius, or arguments groupFile, (haloID | subhaloID), not
        both.

    haloRadius : float
        Aperture within which to select particles around the source
        centre, in simulation units. Provide either arguments
        haloPosition, haloVelocity and haloRadius, or arguments groupFile,
        (haloID | subhaloID), not both.

    groupFile : str
        Path to group file (e.g. /path/to/groups_136). Provide either
        arguments haloPosition, haloVelocity and haloRadius, or arguments
        groupFile, (haloID | subhaloID), not both.

    haloID : int
        ID of FOF group to use as source. Provice either arguments
        haloPosition, haloVelocity and haloRadius, or arguments groupFile,
        (haloID | subhaloID), not both.

    subhaloID : int
        ID of subhalo to use as source. Provide either arguments haloPostion,
        haloVelocity and haloRadius, or arguments groupFile, (haloID |
        subhaloID), not both.

    xH : float
        Primordial hydrogen fraction (default: 0.76).

    Lbox : astropy.units.Quantity, with dimensions of length
        Comoving box side length, without factor h.

    internal_units : dict
        Specify the system of units used in the snapshot file. The dict keys
        should be 'L' (length), 'M' (mass), 'V' (velocity), 'T' (temperature).
        The values should use astropy.units.Quantity. (Default:
        dict(L=U.kpc, M=1E10 * U.Msun, V=U.km/U.s, T=U.K).)

    rescaleRadius : float
        Factor by which to multiply the haloRadius to define the aperture
        within which particles are selected. Useful in conjunction with
        arguments groupFile, (haloID | subhaloID): by default the aperture
        will be the halo virial radius, use this argument to adjust as needed.

    distance : Quantity, with dimensions of length, optional
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: 3 Mpc.)

    vpeculiar : Quantity, with dimensions of velocity, optional
        Source peculiar velocity, added to the velocity from Hubble's law.
        (Default: 0 km/s.)

    rotation : dict, optional
        Must have a single key, which must be one of `axis_angle`, `rotmat` or
        `L_coords`. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - `axis_angle` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - `rotmat` : A (3, 3) numpy.array specifying a rotation.
        - `L_coords` : A 2-tuple containing an inclination and an azimuthal \
        angle (both Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: identity rotation matrix.)

    ra : Quantity, with dimensions of angle, optional
        Right ascension for the source centroid. (Default: 0 deg.)

    dec : Quantity, with dimensions of angle, optional
        Declination for the source centroid. (Default: 0 deg.)
    """

    def __init__(
        self,
        snapBase=None,
        haloPosition=None,
        haloVelocity=None,
        haloRadius=None,
        groupFile=None,
        haloID=None,
        subhaloID=None,
        rescaleRadius=1.0,
        xH=0.76,  # not in header
        Lbox=100 * U.Mpc,  # what is it, actually?
        internal_units=dict(L=U.kpc, M=1e10 * U.Msun, V=U.km / U.s, T=U.K),
        distance=3 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0 * U.deg,
        dec=0 * U.deg,
    ):
        from g3t.stable.g3read import GadgetFile, read_particles_in_box

        # I guess I should allow rescaling of radius to get fore/background

        if (haloID is not None) or (subhaloID is not None) or (groupFile is not None):
            if (
                (haloPosition is not None)
                or (haloVelocity is not None)
                or (haloRadius is not None)
            ):
                raise
        else:
            if (haloID is not None) and (subhaloID is not None):
                raise

        if (haloID is not None) or (subhaloID is not None):
            f = GadgetFile(groupFile)
            data_sub = f.read_new(blocks=["SPOS", "SVEL", "GRNR"], ptypes=[1])
            data_fof = f.read_new(blocks=["RVIR", "FSUB"], ptypes=[0])
            xyz = data_sub["SPOS"]
            vxyz = data_sub["SVEL"]
            fsub = data_fof["FSUB"]
            rvir = data_fof["RVIR"]
            grnr = data_sub["GRNR"]
            if subhaloID is None:
                subhaloID = fsub[haloID]
            if haloID is None:
                haloID = grnr[subhaloID]
            haloPosition = xyz[fsub[haloID]]
            haloVelocity = vxyz[fsub[haloID]]
            haloRadius = rvir[haloID]

        haloRadius *= rescaleRadius

        particles = {}

        # Here all is still in code units
        header = GadgetFile(snapBase + ".0").header

        a = header.time
        h = header.HubbleParam

        l_unit = internal_units["L"] * a / h
        m_unit = internal_units["M"] / h
        v_unit = internal_units["V"] * np.sqrt(a)
        T_unit = internal_units["T"]

        f_gas = read_particles_in_box(
            snapBase,
            haloPosition,
            haloRadius,
            ["POS ", "VEL ", "MASS", "TEMP", "NH  ", "HSML"],
            [0],
        )

        particles["xyz_g"] = f_gas["POS "] * l_unit
        particles["vxyz_g"] = f_gas["VEL "] * v_unit
        particles["hsm_g"] = (
            f_gas["HSML"] * l_unit * find_fwhm(_WendlandC6Kernel().kernel)
        )
        particles["T_g"] = f_gas["TEMP"] * T_unit
        particles["mHI_g"] = f_gas["NH  "] * xH * f_gas["MASS"] * m_unit

        particles["xyz_g"] -= haloPosition * l_unit
        particles["xyz_g"][particles["xyz_g"] > Lbox * a / 2.0] -= Lbox.to(U.kpc) * a
        particles["xyz_g"][particles["xyz_g"] < -Lbox * a / 2.0] += Lbox.to(U.kpc) * a
        particles["vxyz_g"] -= haloVelocity * v_unit

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            **particles,
        )
        return



from . import Martini, DataCube
from .beams import GaussianBeam
from .noise import GaussianNoise
from .spectral_models import GaussianSpectrum
from .sph_kernels import CubicSplineKernel
from .sources import SPHSource
import astropy.units as U
import numpy as np
from scipy.optimize import fsolve


def demo(cubefile="testcube.fits", beamfile="testbeam.fits", hdf5file="testcube.hdf5"):
    """
    Demonstrates basic usage of MARTINI.

    Creates a (very!) crude toy model of a galaxy with a linearly rising
    rotation curve, exponential disk profile, exponential vertical structure. A
    basic configuration of MARTINI is initialized and used to create and output
    a datacube and an image of the beam.

    Parameters
    ----------
    cubefile : string
        File to write demonstration datacube.

    beamfile : string
        File to write demonstration beam.
    """

    # ------make a toy galaxy----------
    N = 500
    phi = np.random.rand(N) * 2 * np.pi
    r = []
    for L in np.random.rand(N):

        def f(r):
            return L - 0.5 * (2 - np.exp(-r) * (np.power(r, 2) + 2 * r + 2))

        r.append(fsolve(f, 1.0)[0])
    r = np.array(r)
    # exponential disk
    r *= 3 / np.sort(r)[N // 2]
    z = -np.log(np.random.rand(N))
    # exponential scale height
    z *= 0.5 / np.sort(z)[N // 2] * np.sign(np.random.rand(N) - 0.5)
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    xyz_g = np.vstack((x, y, z)) * U.kpc
    # linear rotation curve
    vphi = 100 * r / 6.0
    vx = -vphi * np.sin(phi)
    vy = vphi * np.cos(phi)
    # small pure random z velocities
    vz = (np.random.rand(N) * 2.0 - 1.0) * 5
    vxyz_g = np.vstack((vx, vy, vz)) * U.km * U.s**-1
    T_g = np.ones(N) * 8e3 * U.K
    mHI_g = np.ones(N) / N * 5.0e9 * U.Msun
    # ~mean interparticle spacing smoothing
    hsm_g = np.ones(N) * 4 / np.sqrt(N) * U.kpc
    # ---------------------------------

    source = SPHSource(
        distance=3.0 * U.Mpc,
        rotation={"L_coords": (60.0 * U.deg, 0.0 * U.deg)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
        h=0.7,
        T_g=T_g,
        mHI_g=mHI_g,
        xyz_g=xyz_g,
        vxyz_g=vxyz_g,
        hsm_g=hsm_g,
    )

    datacube = DataCube(
        n_px_x=128,
        n_px_y=128,
        n_channels=32,
        px_size=10.0 * U.arcsec,
        channel_width=10.0 * U.km * U.s**-1,
        velocity_centre=source.vsys,
    )

    beam = GaussianBeam(
        bmaj=30.0 * U.arcsec, bmin=30.0 * U.arcsec, bpa=0.0 * U.deg, truncate=4.0
    )

    noise = GaussianNoise(rms=3.0e-5 * U.Jy * U.beam**-1)

    spectral_model = GaussianSpectrum(sigma=7 * U.km * U.s**-1)

    sph_kernel = CubicSplineKernel()

    M = Martini(
        source=source,
        datacube=datacube,
        beam=beam,
        noise=noise,
        spectral_model=spectral_model,
        sph_kernel=sph_kernel,
    )

    M.insert_source_in_cube()
    M.add_noise()
    M.convolve_beam()
    M.write_beam_fits(beamfile, channels="velocity")
    M.write_fits(cubefile, channels="velocity")
    print(f"Wrote demo fits output to {cubefile}, and beam image to {beamfile}.")
    try:
        M.write_hdf5(hdf5file, channels="velocity")
    except ModuleNotFoundError:
        print("h5py package not present, skipping hdf5 output demo.")
    else:
        print(f"Wrote demo hdf5 output to {hdf5file}.")

    return



import numpy as np
import astropy.units as U
from .sph_source import SPHSource


class SOSource(SPHSource):
    """
    Class abstracting HI sources using the SimObj package for interface to
    simulation data.

    This class accesses simulation data via the SimObj package
    (https://github.com/kyleaoman/simobj); see the documentation of that
    package for further configuration instructions.

    Parameters
    ----------
    distance : Quantity, with dimensions of length, optional
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: 3 Mpc.)

    vpeculiar : Quantity, with dimensions of velocity, optional
        Source peculiar velocity, added to the velocity from Hubble's law.
        (Default: 0 km/s.)

    rotation : dict, optional
        Must have a single key, which must be one of `axis_angle`, `rotmat` or
        `L_coords`. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - `axis_angle` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - `rotmat` : A (3, 3) numpy.array specifying a rotation.
        - `L_coords` : A 2-tuple containing an inclination and an azimuthal \
        angle (both Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: identity rotation matrix.)

    ra : Quantity, with dimensions of angle, optional
        Right ascension for the source centroid. (Default: 0 deg.)

    dec : Quantity, with dimensions of angle, optional
        Declination for the source centroid. (Default: 0 deg.)

    SO_args : dict
        Dictionary of keyword arguments to pass to a call to simobj.SimObj.
        Arguments are: `obj_id`, `snap_id`, `mask_type`, `mask_args`,
        `mask_kwargs`, `configfile`, `simfiles_configfile`, `ncpu`. See simobj
        package documentation for details. Provide SO_args or SO_instance, not
        both.

    SO_instance : SimObj instance
        An initialized SimObj object. Provide SO_instance or SO_args, not both.
    """

    def __init__(
        self,
        distance=3.0 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
        SO_args=None,
        SO_instance=None,
        rescale_hsm_g=1,
    ):
        from simobj import SimObj  # optional dependency for this source class

        self._SO_args = SO_args
        self.rescale_hsm_g = rescale_hsm_g
        if (SO_args is not None) and (SO_instance is not None):
            raise ValueError(
                "martini.source.SOSource: Provide SO_args or " "SO_instance, not both."
            )
        elif SO_args is not None:
            with SimObj(**self._SO_args) as SO:
                super().__init__(
                    distance=distance,
                    rotation=rotation,
                    ra=ra,
                    dec=dec,
                    h=SO.h,
                    T_g=SO.T_g,
                    mHI_g=SO.mHI_g,
                    xyz_g=SO.xyz_g,
                    vxyz_g=SO.vxyz_g,
                    hsm_g=SO.hsm_g * self.rescale_hsm_g,
                )
        elif SO_instance is not None:
            super().__init__(
                distance=distance,
                vpeculiar=vpeculiar,
                rotation=rotation,
                ra=ra,
                dec=dec,
                h=SO_instance.h,
                T_g=SO_instance.T_g,
                mHI_g=SO_instance.mHI_g,
                xyz_g=SO_instance.xyz_g,
                vxyz_g=SO_instance.vxyz_g,
                hsm_g=SO_instance.hsm_g,
            )
        else:
            raise ValueError(
                "martini.sources.SOSource: Provide one of SO_args" " or SO_instance."
            )
        return



import numpy as np
from .sph_source import SPHSource
from ..sph_kernels import _WendlandC2Kernel, find_fwhm
from os.path import join, normpath, sep
import astropy.units as U


class EAGLESource(SPHSource):
    """
    Class abstracting HI sources designed to work with publicly available
    EAGLE snapshot + group data.

    For file access, see http://icc.dur.ac.uk/Eagle/database.php.

    Parameters
    ----------
    snapPath : str
        Directory containing snapshot files. The directory structure unpacked
        from the publicly available tarballs is expected; removing/renaming
        files or directories below this will cause errors.

    snapBase : str
        Filename of snapshot files, omitting portion '.X.hdf5'.

    fof : int
        FOF group number of the target object. Note that all particles in the
        FOF group to which the subhalo belongs are used to construct the data
        cube. This avoids strange "holes" at the locations of other subhaloes
        in the same group, and gives a more realistic treatment of foreground
        and background emission local to the source. In the EAGLE database,
        this is the 'GroupNumber'.

    sub : int
        Subfind subhalo number of the target object. For centrals the subhalo
        number is 0, for satellites >0. In the EAGLE database, this is then
        'SubGroupNumber'.

    db_user : str
        Database username.

    db_key : str, optional
        Database password, or omit for a prompt at runtime. (Default: None.)

    subBoxSize : Quantity, with dimensions of length
        Box half-side length of a region to load around the object of interest,
        in physical (not comoving, no little h) units. Using larger values
        will include more foreground/background, which may be desirable, but
        will also slow down execution and impair the automatic routine used
        to find a disc plane.

    distance : Quantity, with dimensions of length, optional
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: 3 Mpc.)

    vpeculiar : Quantity, with dimensions of velocity, optional
        Source peculiar velocity, added to the velocity from Hubble's law.
        (Default: 0 km/s.)

    rotation : dict, optional
        Must have a single key, which must be one of `axis_angle`, `rotmat` or
        `L_coords`. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - `axis_angle` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - `rotmat` : A (3, 3) numpy.array specifying a rotation.
        - `L_coords` : A 2-tuple containing an inclination and an azimuthal \
        angle (both Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: identity rotation matrix.)

    ra : Quantity, with dimensions of angle, optional
        Right ascension for the source centroid. (Default: 0 deg.)

    dec : Quantity, with dimensions of angle, optional
        Declination for the source centroid. (Default: 0 deg.)

    print_query : bool, optional
        If True, the SQL query submitted to the EAGLE database is printed.
        (Default: False.)
    """

    def __init__(
        self,
        snapPath=None,
        snapBase=None,
        fof=None,
        sub=None,
        db_user=None,
        db_key=None,
        subBoxSize=50.0 * U.kpc,
        distance=3.0 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
        print_query=False,
    ):
        if snapPath is None:
            raise ValueError("Provide snapPath argument to EAGLESource.")
        if snapBase is None:
            raise ValueError("Provide snapBase argument to EAGLESource.")
        if fof is None:
            raise ValueError("Provide fof argument to EAGLESource.")
        if sub is None:
            raise ValueError("Provide sub argument to EAGLESource.")
        if db_user is None:
            raise ValueError("Provide EAGLE database username.")

        # optional dependencies for this source class
        from eagleSqlTools import connect, execute_query
        from pyread_eagle import EagleSnapshot
        from Hdecompose.atomic_frac import atomic_frac
        import h5py

        snapNum = int(snapBase.split("_")[1])
        volCode = normpath(snapPath).split(sep)[-2]
        query = (
            "SELECT "
            "  sh.redshift as redshift, "
            "  sh.CentreOfPotential_x as x, "
            "  sh.CentreOfPotential_y as y, "
            "  sh.CentreOfPotential_z as z, "
            "  sh.Velocity_x as vx, "
            "  sh.Velocity_y as vy, "
            "  sh.Velocity_z as vz "
            "FROM "
            "  {:s}_SubHalo as sh ".format(volCode) + "WHERE "
            "  sh.Snapnum = {:d} ".format(snapNum)
            + "  and sh.GroupNumber = {:d} ".format(fof)
            + "  and sh.SubGroupNumber = {:d}".format(sub)
        )
        if print_query:
            print("-----EAGLE-DB-QUERY-----")
            print(query)
            print("-------QUERY-ENDS-------")
        if db_key is None:
            print("EAGLE database")
        q = execute_query(connect(db_user, db_key), query)
        redshift = q["redshift"]
        a = np.power(1 + redshift, -1)
        cop = np.array([q[coord] for coord in "xyz"]) * a * U.Mpc
        vcent = np.array([q["v" + coord] for coord in "xyz"]) * U.km / U.s

        snapFile = join(snapPath, snapBase + ".0.hdf5")

        with h5py.File(snapFile, "r") as f:
            h = f["RuntimePars"].attrs["HubbleParam"]
            subBoxSize = (subBoxSize * h / a).to(U.Mpc).value
            centre = (cop * h / a).to(U.Mpc).value
            eagle_data = EagleSnapshot(snapFile)
            region = np.vstack((centre - subBoxSize, centre + subBoxSize)).T.flatten()
            eagle_data.select_region(*region)
            lbox = f["/Header"].attrs["BoxSize"] * U.Mpc / h
            fH = f["/RuntimePars"].attrs["InitAbundance_Hydrogen"]
            fHe = f["/RuntimePars"].attrs["InitAbundance_Helium"]
            proton_mass = f["/Constants"].attrs["PROTONMASS"] * U.g
            mu = 1 / (fH + 0.25 * fHe)
            gamma = f["/RuntimePars"].attrs["EOS_Jeans_GammaEffective"]
            T0 = f["/RuntimePars"].attrs["EOS_Jeans_TempNorm_K"] * U.K

            def fetch(att, ptype=0):
                # gas is type 0, only need gas properties
                tmp = eagle_data.read_dataset(ptype, att)
                dset = f["/PartType{:d}/{:s}".format(ptype, att)]
                aexp = dset.attrs.get("aexp-scale-exponent")
                hexp = dset.attrs.get("h-scale-exponent")
                return tmp[()] * np.power(a, aexp) * np.power(h, hexp)

            code_to_g = f["/Units"].attrs["UnitMass_in_g"] * U.g
            code_to_cm = f["/Units"].attrs["UnitLength_in_cm"] * U.cm
            code_to_cm_s = f["/Units"].attrs["UnitVelocity_in_cm_per_s"] * U.cm / U.s
            ng_g = fetch("GroupNumber")
            particles = dict(
                xyz_g=(fetch("Coordinates") * code_to_cm).to(U.kpc),
                vxyz_g=(fetch("Velocity") * code_to_cm_s).to(U.km / U.s),
                T_g=fetch("Temperature") * U.K,
                hsm_g=(fetch("SmoothingLength") * code_to_cm).to(U.kpc)
                * find_fwhm(_WendlandC2Kernel().kernel),
            )
            rho_g = fetch("Density") * U.g * U.cm**-3
            SFR_g = fetch("StarFormationRate")
            Habundance_g = fetch("ElementAbundance/Hydrogen")

        particles["mHI_g"] = (
            atomic_frac(
                redshift,
                rho_g * Habundance_g / (mu * proton_mass),
                particles["T_g"],
                rho_g,
                Habundance_g,
                onlyA1=True,
                EAGLE_corrections=True,
                SFR=SFR_g,
                mu=mu,
                gamma=gamma,
                fH=fH,
                T0=T0,
            )
            * code_to_g
        ).to(U.Msun)

        mask = ng_g == fof
        for k, v in particles.items():
            particles[k] = v[mask]

        particles["xyz_g"] -= cop
        particles["xyz_g"][particles["xyz_g"] > lbox / 2.0] -= lbox.to(U.kpc)
        particles["xyz_g"][particles["xyz_g"] < -lbox / 2.0] += lbox.to(U.kpc)
        particles["vxyz_g"] -= vcent

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            **particles,
        )
        return



import os
import pytest
import numpy as np
from astropy import units as U
from astropy.coordinates.matrix_utilities import rotation_matrix
from martini.datacube import DataCube
from martini.sources import SPHSource
from martini.sources._cartesian_translation import translate, translate_d
from martini.sources._L_align import L_align
from astropy.coordinates import CartesianRepresentation, CartesianDifferential


class TestSourceUtilities:
    def test_L_align(self):
        """
        Test that L_align produces expected rotation matrices, including saving to file.
        """
        # set up 4 particles in x-y plane rotating right-handed about zhat
        _xyz = np.array([[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0]])
        _vxyz = np.array([[0, 1, 0], [-1, 0, 0], [0, -1, 0], [1, 0, 0]])
        # tile outwards to make a 4-armed "windmill"
        xyz = np.vstack((_xyz * np.arange(1, 201).reshape(200, 1, 1))) * U.kpc
        vxyz = np.vstack((_vxyz * np.arange(1, 201).reshape(200, 1, 1))) * U.km / U.s
        m = np.ones(xyz.shape[0]) * U.Msun
        # rotating zhat to align with zhat should stay aligned with zhat
        # (but might arbitrarily rotate x-y plane)
        assert np.allclose(
            np.array([[0, 0, 1]]).dot(L_align(xyz, vxyz, m, Laxis="z")),
            np.array([0, 0, 1]),
        )
        # rotating zhat to align with xhat should align with xhat
        assert np.allclose(
            np.array([[0, 0, 1]]).dot(L_align(xyz, vxyz, m, Laxis="x")),
            np.array([1, 0, 0]),
        )
        # rotating zhat to align with yhat should align with yhat
        assert np.allclose(
            np.array([[0, 0, 1]]).dot(L_align(xyz, vxyz, m, Laxis="y")),
            np.array([0, 1, 0]),
        )
        # and also check transposed cases
        assert np.allclose(
            L_align(xyz.T, vxyz.T, m, Laxis="z").dot(np.array([0, 0, 1])),
            np.array([0, 0, 1]),
        )
        assert np.allclose(
            L_align(xyz.T, vxyz.T, m, Laxis="x").dot(np.array([0, 0, 1])),
            np.array([1, 0, 0]),
        )
        assert np.allclose(
            L_align(xyz.T, vxyz.T, m, Laxis="y").dot(np.array([0, 0, 1])),
            np.array([0, 1, 0]),
        )
        # finally check that saving to file works
        testfile = "testsaverot.npy"
        try:
            rotmat = L_align(xyz, vxyz, m, saverot=testfile)
            saved_rotmat = np.load(testfile)
        finally:
            if os.path.exists(testfile):
                os.remove(testfile)
        assert np.allclose(rotmat, saved_rotmat)

    def test_translate(self):
        """
        Check that cartesian representation transforms correctly.
        """
        setattr(CartesianRepresentation, "translate", translate)
        cr = CartesianRepresentation(np.zeros(3) * U.kpc)
        translation = np.ones(3) * U.kpc
        cr_translated = cr.translate(translation)
        assert U.allclose(cr_translated.get_xyz(), translation)

    def test_translate_d(self):
        """
        Check that cartesian differential transforms correctly.
        """
        setattr(CartesianDifferential, "translate", translate_d)
        cd = CartesianDifferential(np.zeros(3) * U.km / U.s)
        translation = np.ones(3) * U.km / U.s
        cd_translated = cd.translate(translation)
        assert U.allclose(cd_translated.get_d_xyz(), translation)


class TestSPHSource:
    def test_coordinate_input(self):
        """
        Check that different input shapes for coordinates give expected behaviour.
        """
        mHI_g = np.zeros(4) * U.Msun
        row_coords = np.zeros((4, 3)) * U.kpc
        row_vels = np.zeros((4, 3)) * U.km / U.s
        s = SPHSource(xyz_g=row_coords, vxyz_g=row_vels, mHI_g=mHI_g)
        assert s.coordinates_g.shape == (4,)
        assert s.coordinates_g.differentials["s"].shape == (4,)
        assert s.npart == 4
        col_coords = np.zeros((3, 4)) * U.kpc
        col_vels = np.zeros((3, 4)) * U.km / U.s
        s = SPHSource(xyz_g=col_coords, vxyz_g=col_vels, mHI_g=mHI_g)
        assert s.coordinates_g.shape == (4,)
        assert s.coordinates_g.differentials["s"].shape == (4,)
        assert s.npart == 4
        mHI_g = np.zeros(3) * U.Msun
        symm_coords = np.zeros((3, 3)) * U.kpc
        symm_coords[0, 1] = 1 * U.kpc
        symm_vels = np.zeros((3, 3)) * U.km / U.s
        symm_vels[0, 1] = 1 * U.km / U.s
        s = SPHSource(
            xyz_g=symm_coords, vxyz_g=symm_vels, mHI_g=mHI_g, coordinate_axis=0
        )
        assert s.coordinates_g.shape == (3,)
        assert s.coordinates_g.differentials["s"].shape == (3,)
        assert s.coordinates_g.x[1] == 1 * U.kpc
        assert s.coordinates_g.differentials["s"].d_x[1] == 1 * U.km / U.s
        assert s.npart == 3
        s = SPHSource(
            xyz_g=symm_coords, vxyz_g=symm_vels, mHI_g=mHI_g, coordinate_axis=1
        )
        assert s.coordinates_g.shape == (3,)
        assert s.coordinates_g.shape == (3,)
        assert s.coordinates_g.y[0] == 1 * U.kpc
        assert s.coordinates_g.differentials["s"].d_y[0] == 1 * U.km / U.s
        assert s.npart == 3
        with pytest.raises(RuntimeError, match="cannot guess coordinate_axis"):
            SPHSource(xyz_g=symm_coords, vxyz_g=symm_vels, mHI_g=mHI_g)
        with pytest.raises(ValueError, match="must have matching shapes"):
            SPHSource(xyz_g=row_coords, vxyz_g=col_vels, mHI_g=mHI_g)

    @pytest.mark.parametrize("ra", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("dec", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    def test_ra_dec_rotation(self, ra, dec):
        """
        Check that coordinates rotate as needed before translation to observed position.
        """
        xyz_g = np.arange(3).reshape((1, 3)) * U.kpc
        vxyz_g = np.arange(3).reshape((1, 3)) * U.km / U.s
        mHI_g = np.zeros(3) * U.Msun
        s = SPHSource(
            xyz_g=xyz_g, vxyz_g=vxyz_g, mHI_g=mHI_g, ra=ra, dec=dec, distance=0 * U.Mpc
        )
        s._init_skycoords(_reset=False)
        R_y = np.array(
            [
                [np.cos(dec), 0, np.sin(dec)],
                [0, 1, 0],
                [-np.sin(dec), 0, np.cos(dec)],
            ]
        )
        R_z = np.array(
            [
                [np.cos(-ra), -np.sin(-ra), 0],
                [np.sin(-ra), np.cos(-ra), 0],
                [0, 0, 1],
            ]
        )
        rotmat = R_y.dot(R_z)
        assert U.allclose(s.coordinates_g.xyz.T, xyz_g.dot(rotmat))
        assert U.allclose(
            s.coordinates_g.differentials["s"].d_xyz.T, vxyz_g.dot(rotmat)
        )

    @pytest.mark.parametrize("ra", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("dec", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("distance", (0 * U.Mpc, 3 * U.Mpc))
    @pytest.mark.parametrize("vpeculiar", (0 * U.km / U.s, 100 * U.km / U.s))
    def test_dist_vpec_translation(self, ra, dec, distance, vpeculiar):
        """
        Check that coordinates translate correctly to observed position.
        """
        xyz_g = np.arange(3).reshape((1, 3)) * U.kpc
        vxyz_g = np.arange(3).reshape((1, 3)) * U.km / U.s
        mHI_g = np.zeros(3) * U.Msun
        s = SPHSource(
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            mHI_g=mHI_g,
            ra=ra,
            dec=dec,
            distance=distance,
            vpeculiar=vpeculiar,
        )
        s._init_skycoords(_reset=False)
        vsys = s.h * 100 * U.km / U.s / U.Mpc * distance + vpeculiar
        R_y = np.array(
            [
                [np.cos(dec), 0, np.sin(dec)],
                [0, 1, 0],
                [-np.sin(dec), 0, np.cos(dec)],
            ]
        )
        R_z = np.array(
            [
                [np.cos(-ra), -np.sin(-ra), 0],
                [np.sin(-ra), np.cos(-ra), 0],
                [0, 0, 1],
            ]
        )
        rotmat = R_y.dot(R_z)
        direction_vector = rotmat.T.dot(np.array([[1], [0], [0]]))
        assert U.allclose(
            s.coordinates_g.xyz.T, xyz_g.dot(rotmat) + direction_vector.T * distance
        )
        assert U.allclose(
            s.coordinates_g.differentials["s"].d_xyz.T,
            vxyz_g.dot(rotmat) + direction_vector.T * vsys,
        )

    def test_init_skycoords_resets(self, s):
        """
        Check that particle coordinate arrays are reset after initialising skycoords.
        """
        initial_coords = s.coordinates_g
        s._init_skycoords()
        ax_equal = [
            U.allclose(getattr(s.coordinates_g, ax), getattr(initial_coords, ax))
            for ax in "xyz"
        ]
        assert all(ax_equal)
        # make sure that it wasn't a no-op:
        s._init_skycoords(_reset=False)
        ax_equal = [
            U.allclose(getattr(s.coordinates_g, ax), getattr(initial_coords, ax))
            for ax in "xyz"
        ]
        assert not all(ax_equal)

    def test_init_pixcoords(self):
        """
        Check that pixel coordinates are accurately calculated from angular positions and
        velocity offsets.
        """
        # set distance so that 1kpc = 1arcsec
        distance = (1 * U.kpc / 1 / U.arcsec).to(U.Mpc, U.dimensionless_angles())
        # line up particles 1 per 1kpc = 1arcsec interval in RA and Dec
        # and 1 per 1 km / s interval in vlos
        # set h=0 so that velocity stays centred at 0
        source = SPHSource(
            distance=distance,
            h=0.0,
            T_g=np.ones(5) * 1e4 * U.K,
            mHI_g=np.ones(5) * 1e4 * U.Msun,
            xyz_g=U.Quantity(
                np.vstack(
                    (
                        np.zeros(6),
                        np.linspace(-2.5, 2.5, 6),
                        (np.linspace(-2.5, 2.5, 6)),
                    )
                ).T,
                U.kpc,
            ),
            vxyz_g=U.Quantity(
                np.vstack((np.linspace(-2.5, 2.5, 6), np.zeros(6), np.zeros(6))).T,
                U.km / U.s,
            ),
            hsm_g=np.ones(6) * U.kpc,
        )
        datacube = DataCube(
            n_px_x=6,
            n_px_y=6,
            n_channels=6,
            px_size=1 * U.arcsec,
            channel_width=1 * U.km / U.s,
        )
        expected_coords = (
            np.vstack((np.arange(6)[::-1], np.arange(6), np.arange(6))) * U.pix
        )
        source._init_skycoords()
        source._init_pixcoords(datacube)
        assert U.allclose(
            source.pixcoords,
            expected_coords,
            atol=1e-4 * U.pix,
        )

    @pytest.mark.parametrize("ra", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("dec", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    def test_sky_location(self, ra, dec):
        """
        Check that particle ends up where expected on the sky.
        """
        from astropy.coordinates import Angle

        xyz_g = np.zeros((1, 3)) * U.kpc
        vxyz_g = np.zeros((1, 3)) * U.km / U.s
        mHI_g = np.zeros(1) * U.Msun
        s = SPHSource(xyz_g=xyz_g, vxyz_g=vxyz_g, ra=ra, dec=dec, mHI_g=mHI_g)
        s._init_skycoords()
        assert U.isclose(s.skycoords.ra[0], Angle(ra).wrap_at(360 * U.deg))
        assert U.isclose(s.skycoords.dec[0], Angle(dec).wrap_at(180 * U.deg))

    def test_apply_mask(self, s):
        """
        Check that particle arrays can be masked.
        """
        particle_fields = ("T_g", "mHI_g", "hsm_g")
        npart_before_mask = s.T_g.size
        particles_before = {k: getattr(s, k) for k in particle_fields}
        particles_before["coordinates_g"] = s.coordinates_g
        # make sure we have a scalar value for one field to check it's handled
        assert particles_before["hsm_g"].isscalar
        mask = np.r_[
            np.ones(npart_before_mask - npart_before_mask // 3, dtype=int),
            np.zeros(npart_before_mask // 3, dtype=int),
        ]
        s.apply_mask(mask)
        for k in particle_fields:
            if not particles_before[k].isscalar:
                assert U.allclose(particles_before[k][mask], getattr(s, k))
            else:
                assert U.isclose(particles_before[k], getattr(s, k))
        assert U.allclose(
            particles_before["coordinates_g"].get_xyz()[:, mask],
            s.coordinates_g.get_xyz(),
        )
        assert U.allclose(
            particles_before["coordinates_g"].differentials["s"].get_d_xyz()[:, mask],
            s.coordinates_g.differentials["s"].get_d_xyz(),
        )

    def test_apply_badmask(self, s):
        """
        Check that bad masks are rejected.
        """
        with pytest.raises(
            ValueError, match="Mask must have same length as particle arrays."
        ):
            s.apply_mask(np.array([1, 2, 3]))
        with pytest.raises(RuntimeError, match="No source particles in target region."):
            s.apply_mask(np.zeros(s.npart, dtype=int))

    def test_rotate_axis_angle(self, s):
        """
        Test that we can rotate by an axis-angle transformation.
        """
        assert np.allclose(s.current_rotation, np.eye(3))
        axis = "z"
        angle = 30 * U.deg
        # expect a right-handed rotation:
        rotmat = np.array(
            [
                [np.cos(angle), -np.sin(angle), 0],
                [np.sin(angle), np.cos(angle), 0],
                [0, 0, 1],
            ]
        )
        s.rotate(axis_angle=(axis, angle))
        assert np.allclose(s.current_rotation, rotmat)

    def test_rotate_rotmat(self, s):
        """
        Check that we can rotate by a rotmat transformation.
        """
        assert np.allclose(s.current_rotation, np.eye(3))
        angle = 30 * U.deg
        rotmat = np.array(
            [
                [np.cos(angle), -np.sin(angle), 0],
                [np.sin(angle), np.cos(angle), 0],
                [0, 0, 1],
            ]
        )
        vector_before = s.coordinates_g.get_xyz()[:, 0]
        assert any(np.abs(vector_before) > 0)
        s.rotate(rotmat=rotmat)
        assert U.allclose(
            s.coordinates_g.get_xyz()[:, 0], np.dot(rotmat, vector_before)
        )

    @pytest.mark.parametrize("incl", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("az_rot", (0 * U.deg, 30 * U.deg, -30 * U.deg))
    @pytest.mark.parametrize("pa", (270 * U.deg, 300 * U.deg, 240 * U.deg))
    def test_rotate_L_coords(self, s, incl, az_rot, pa):
        """
        Check that we can rotate automatically to the angular momentum frame.
        """
        assert np.allclose(s.current_rotation, np.eye(3))
        rotmat = L_align(
            s.coordinates_g.get_xyz(),
            s.coordinates_g.differentials["s"].get_d_xyz(),
            s.mHI_g,
            Laxis="x",
        )
        rotmat = rotation_matrix(az_rot, axis="x").T.dot(rotmat)
        rotmat = rotation_matrix(incl, axis="y").T.dot(rotmat)
        if incl >= 0:
            rotmat = rotation_matrix(pa - 90 * U.deg, axis="x").T.dot(rotmat)
        else:
            rotmat = rotation_matrix(pa - 270 * U.deg, axis="x").T.dot(rotmat)
        if U.isclose(pa, 270 * U.deg):
            s.rotate(L_coords=(incl, az_rot))
        else:
            s.rotate(L_coords=(incl, az_rot, pa))
        assert np.allclose(s.current_rotation, rotmat)

    def test_composite_rotations(self, s):
        """
        Check that multiple rotations in a function call are blocked.
        """
        with pytest.raises(
            ValueError, match="Multiple rotations in a single call not allowed."
        ):
            s.rotate(axis_angle=("x", 30 * U.deg), rotmat=np.eye(3))

    def test_translate(self, s):
        """
        Check that coordinates translate correctly.
        """
        for translation_shape in ((3, 1), (1, 3)):
            initial_coords = s.coordinates_g.get_xyz()
            translation = np.ones(3) * U.kpc
            s.translate(translation.reshape(translation_shape))
            expected_coords = initial_coords + translation.reshape((3, 1))
            assert U.allclose(s.coordinates_g.get_xyz(), expected_coords)

    def test_boost(self, s):
        """
        Check that velocities translate correctly.
        """
        for translation_shape in ((3, 1), (1, 3)):
            initial_vels = s.coordinates_g.differentials["s"].get_d_xyz()
            translation = np.ones(3) * U.km / U.s
            s.boost(translation.reshape(translation_shape))
            expected_vels = initial_vels + translation.reshape((3, 1))
            assert U.allclose(
                s.coordinates_g.differentials["s"].get_d_xyz(), expected_vels
            )

    def test_save_current_rotation(self, s):
        """
        Check that current rotation state can be output to file.
        """
        assert np.allclose(s.current_rotation, np.eye(3))
        angle = np.pi / 4
        rotmat = np.array(
            [
                [np.cos(angle), -np.sin(angle), 0],
                [np.sin(angle), np.cos(angle), 0],
                [0, 0, 1],
            ]
        )
        s.rotate(rotmat=rotmat)
        testfile = "testrotmat.npy"
        try:
            s.save_current_rotation(testfile)
            saved_rotmat = np.loadtxt(testfile)
        finally:
            if os.path.exists(testfile):
                os.remove(testfile)
        assert np.allclose(saved_rotmat, rotmat)


class TestSOSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestSWIFTGalaxySource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestColibreSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestEagleSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestMagneticumSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestSimbaSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError


class TestTNGSource:
    @pytest.mark.xfail
    def test_stuff(self):
        raise NotImplementedError



import numpy as np
from .sph_source import SPHSource
from ..sph_kernels import _CubicSplineKernel, find_fwhm
from os.path import join
from astropy import units as U, constants as C


class SimbaSource(SPHSource):
    """
    Class abstracting HI sources designed to work with SIMBA snapshot and
    group data.

    For file access, enquire with R. Davé (rad@roe.ac.uk).

    Parameters
    ----------
    snapPath : str
        Directory containing snapshot files.

    snapName : str
        Filename of snapshot file.

    groupPath : str
        Directory containing group catalogue files.

    groupName : str
        Group catalogue filename.

    groupID : int
        Identifier in the GroupID column of group catalogue.

    aperture : Quantity, with dimensions of length
        Radial extent of a region to load around the object of interest,
        in physical (not comoving, no little h) units. Using larger values
        will include more foreground/background, which may be desirable, but
        will also slow down execution and impair the automatic routine used
        to find a disc plane.

    distance : Quantity, with dimensions of length, optional
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: 3 Mpc.)

    vpeculiar : Quantity, with dimensions of velocity, optional
        Source peculiar velocity, added to the velocity from Hubble's law.
        (Default: 0 km/s.)

    rotation : dict, optional
        Must have a single key, which must be one of `axis_angle`, `rotmat` or
        `L_coords`. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - `axis_angle` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - `rotmat` : A (3, 3) numpy.array specifying a rotation.
        - `L_coords` : A 2-tuple containing an inclination and an azimuthal \
        angle (both Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: identity rotation matrix.)

    ra : Quantity, with dimensions of angle, optional
        Right ascension for the source centroid. (Default: 0 deg.)

    dec : Quantity, with dimensions of angle, optional
        Declination for the source centroid. (Default: 0 deg.)

    """

    def __init__(
        self,
        snapPath=None,
        snapName=None,
        groupPath=None,
        groupName=None,
        groupID=None,
        aperture=50.0 * U.kpc,
        distance=3.0 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
    ):
        if snapPath is None:
            raise ValueError("Provide snapPath argument to SimbaSource.")
        if snapName is None:
            raise ValueError("Provide snapName argument to SimbaSource.")
        if groupPath is None:
            raise ValueError("Provide groupPath argument to SimbaSource.")
        if groupName is None:
            raise ValueError("Provide groupName argument to SimbaSource.")
        if groupID is None:
            raise ValueError("Provide groupID argument to SimbaSource.")

        # optional dependencies for this source class
        import h5py

        snapFile = join(snapPath, snapName)
        groupFile = join(groupPath, groupName)

        gamma = 5 / 3

        with h5py.File(snapFile, "r") as f:
            a = f["Header"].attrs["Time"]
            h = f["Header"].attrs["HubbleParam"]
            lbox = f["Header"].attrs["BoxSize"] / h * U.kpc
            gas = f["PartType0"]
            fZ = gas["Metallicity"][()][:, 0]
            fHe = gas["Metallicity"][()][:, 1]
            fH = 1 - fHe - fZ
            xe = gas["ElectronAbundance"][()]
            particles = dict(
                xyz_g=gas["Coordinates"][()] * a / h * U.kpc,
                vxyz_g=gas["Velocities"][()] * np.sqrt(a) * U.km / U.s,
                T_g=(
                    (4 / (1 + 3 * fH + 4 * fH * xe))
                    * C.m_p
                    * (gamma - 1)
                    * gas["InternalEnergy"][()]
                    * (U.km / U.s) ** 2
                    / C.k_B
                ).to(U.K),
                hsm_g=gas["SmoothingLength"][()]
                * a
                / h
                * U.kpc
                * find_fwhm(_CubicSplineKernel().kernel),
                mHI_g=gas["Masses"][()] * fH * gas["GrackleHI"][()] * 1e10 / h * U.Msun,
            )
            del fH, fHe, xe

        with h5py.File(groupFile, "r") as f:
            groupIDs = f["galaxy_data/GroupID"][()]
            gmask = groupID == groupIDs
            # no h^-1 on minpotpos, not sure about comoving yet
            cop = f["galaxy_data/minpotpos"][()][gmask][0] * a * U.kpc
            vcent = f["galaxy_data/vel"][()][gmask][0] * np.sqrt(a) * U.km / U.s

        particles["xyz_g"] -= cop
        particles["xyz_g"][particles["xyz_g"] > lbox / 2.0] -= lbox
        particles["xyz_g"][particles["xyz_g"] < -lbox / 2.0] += lbox
        particles["vxyz_g"] -= vcent

        mask = np.zeros(particles["xyz_g"].shape[0], dtype=bool)
        outer_cube = (np.abs(particles["xyz_g"]) < aperture).all(axis=1)
        inner_cube = np.zeros(particles["xyz_g"].shape[0], dtype=bool)
        inner_cube[outer_cube] = (
            np.abs(particles["xyz_g"][outer_cube]) < aperture / np.sqrt(3)
        ).all(axis=1)
        need_distance = np.logical_and(outer_cube, np.logical_not(inner_cube))
        mask[inner_cube] = True
        mask[need_distance] = np.sum(
            np.power(particles["xyz_g"][need_distance], 2), axis=1
        ) < np.power(aperture, 2)

        for k, v in particles.items():
            particles[k] = v[mask]

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            **particles,
        )
        return
